use anyhow::{anyhow, bail};
use chrono::{DateTime, Utc};
use ds18b20::Ds18b20;
use ds323x::{DateTimeAccess, Ds323x};
use eeprom24x::{Eeprom24x, Eeprom24xTrait, SlaveAddr};
use embedded_hal::digital::OutputPin;
use embedded_hal_bus::i2c::MutexDevice;
use esp_idf_hal::adc::oneshot::{AdcChannelDriver, AdcDriver};
use esp_idf_hal::delay::Delay;
use esp_idf_hal::gpio::{AnyInputPin, Gpio5, IOPin, InputOutput, Output, PinDriver, Pull};
use esp_idf_hal::i2c::I2cDriver;
use esp_idf_hal::pcnt::{PcntChannel, PcntChannelConfig, PcntControlMode, PcntCountMode, PcntDriver, PinIndex};
use esp_idf_sys::{gpio_hold_dis, gpio_hold_en, vTaskDelay, EspError};
use one_wire_bus::OneWire;
use pca9535::{GPIOBank, Pca9535Immediate, StandardExpanderInterface};
use crate::hal::{deep_sleep, BackupHeader, BoardInteraction, FreePeripherals, Sensor, I2C_DRIVER, PLANT_COUNT, REPEAT_MOIST_MEASURE, TANK_MULTI_SAMPLE, X25};
use crate::hal::esp::ESP;
use std::result::Result::Ok as OkStd;
use esp_idf_hal::adc::{attenuation, Resolution};
use esp_idf_hal::adc::oneshot::config::AdcChannelConfig;
use crate::config::PlantControllerConfig;
use crate::hal::battery::{BatteryInteraction, BatteryMonitor};
use crate::log::{log, LogMessage};

const MS0: u8 = 1_u8;
const MS1: u8 = 0_u8;
const MS2: u8 = 3_u8;
const MS3: u8 = 4_u8;
const MS4: u8 = 2_u8;
const SENSOR_ON: u8 = 5_u8;

pub struct V4<'a> {
    esp: ESP<'a>,
    battery_monitor: Box<dyn BatteryInteraction>,
    config: PlantControllerConfig,
    tank_channel: AdcChannelDriver<'a, Gpio5, AdcDriver<'a, esp_idf_hal::adc::ADC1>>,
    solar_is_day: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, esp_idf_hal::gpio::Input>,
    signal_counter: PcntDriver<'a>,
    charge_indicator: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
    awake: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, Output>,
    light: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
    tank_power: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
    one_wire_bus: OneWire<PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>>,
    rtc:
        Ds323x<ds323x::interface::I2cInterface<MutexDevice<'a, I2cDriver<'a>>>, ds323x::ic::DS3231>,
    eeprom: Eeprom24x<
        MutexDevice<'a, I2cDriver<'a>>,
        eeprom24x::page_size::B32,
        eeprom24x::addr_size::TwoBytes,
        eeprom24x::unique_serial::No,
    >,
    general_fault: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
    pump_expander: Pca9535Immediate<MutexDevice<'a, I2cDriver<'a>>>,
    sensor_expander: Pca9535Immediate<MutexDevice<'a, I2cDriver<'a>>>,
}


pub(crate) fn create_v4(peripherals: FreePeripherals, esp:ESP, config:PlantControllerConfig, battery_monitor: Box<dyn BatteryInteraction>) -> anyhow::Result<Box<dyn BoardInteraction + '_>> {
    let mut awake = PinDriver::output(peripherals.gpio15.downgrade())?;
    awake.set_high()?;

    let mut general_fault = PinDriver::input_output(peripherals.gpio6.downgrade())?;
    general_fault.set_pull(Pull::Floating)?;
    general_fault.set_low()?;

    println!("Init rtc driver");
    let mut rtc = Ds323x::new_ds3231(MutexDevice::new(&I2C_DRIVER));

    println!("Init rtc eeprom driver");
    let mut eeprom = {
        Eeprom24x::new_24x32(
            MutexDevice::new(&I2C_DRIVER),
            SlaveAddr::Alternative(true, true, true),
        )
    };

    let mut one_wire_pin = PinDriver::input_output_od(peripherals.gpio18.downgrade())?;
    one_wire_pin.set_pull(Pull::Floating)?;

    let one_wire_bus = OneWire::new(one_wire_pin)
        .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;

    let rtc_time = rtc.datetime();
    match rtc_time {
        OkStd(tt) => {
            println!("Rtc Module reports time at UTC {}", tt);
        }
        Err(err) => {
            println!("Rtc Module could not be read {:?}", err);
        }
    }
    match eeprom.read_byte(0) {
        OkStd(byte) => {
            println!("Read first byte with status {}", byte);
        }
        Err(err) => {
            println!("Eeprom could not read first byte {:?}", err);
        }
    }

    let mut signal_counter = PcntDriver::new(
        peripherals.pcnt0,
        Some(peripherals.gpio22),
        Option::<AnyInputPin>::None,
        Option::<AnyInputPin>::None,
        Option::<AnyInputPin>::None,
    )?;

    signal_counter.channel_config(
        PcntChannel::Channel0,
        PinIndex::Pin0,
        PinIndex::Pin1,
        &PcntChannelConfig {
            lctrl_mode: PcntControlMode::Keep,
            hctrl_mode: PcntControlMode::Keep,
            pos_mode: PcntCountMode::Increment,
            neg_mode: PcntCountMode::Hold,
            counter_h_lim: i16::MAX,
            counter_l_lim: 0,
        },
    )?;

    let adc_config = AdcChannelConfig {
        attenuation: attenuation::DB_11,
        resolution: Resolution::Resolution12Bit,
        calibration: esp_idf_hal::adc::oneshot::config::Calibration::Curve,
    };
    let tank_driver = AdcDriver::new(peripherals.adc1)?;
    let tank_channel: AdcChannelDriver<Gpio5, AdcDriver<esp_idf_hal::adc::ADC1>> =
        AdcChannelDriver::new(tank_driver, peripherals.gpio5, &adc_config)?;

    let mut solar_is_day = PinDriver::input(peripherals.gpio7.downgrade())?;
    solar_is_day.set_pull(Pull::Floating)?;

    let mut light = PinDriver::input_output(peripherals.gpio10.downgrade())?;
    light.set_pull(Pull::Floating)?;

    let mut tank_power = PinDriver::input_output(peripherals.gpio11.downgrade())?;
    tank_power.set_pull(Pull::Floating)?;

    let mut charge_indicator = PinDriver::input_output(peripherals.gpio3.downgrade())?;
    charge_indicator.set_pull(Pull::Floating)?;
    charge_indicator.set_low()?;

    let mut pump_expander = Pca9535Immediate::new(MutexDevice::new(&I2C_DRIVER), 32);

    //todo error handing if init error
    for pin in 0..8 {
        let _ = pump_expander.pin_into_output(GPIOBank::Bank0, pin);
        let _ = pump_expander.pin_into_output(GPIOBank::Bank1, pin);
        let _ = pump_expander.pin_set_low(GPIOBank::Bank0, pin);
        let _ = pump_expander.pin_set_low(GPIOBank::Bank1, pin);
    }

    let mut sensor_expander = Pca9535Immediate::new(MutexDevice::new(&I2C_DRIVER), 34);
    for pin in 0..8 {
        let _ = sensor_expander.pin_into_output(GPIOBank::Bank0, pin);
        let _ = sensor_expander.pin_into_output(GPIOBank::Bank1, pin);
        let _ = sensor_expander.pin_set_low(GPIOBank::Bank0, pin);
        let _ = sensor_expander.pin_set_low(GPIOBank::Bank1, pin);
    }

    let v = V4 {
        esp,
        awake,
        tank_channel,
        solar_is_day,
        signal_counter,
        light,
        tank_power,
        one_wire_bus,
        rtc,
        eeprom,
        general_fault,
        pump_expander,
        sensor_expander,
        charge_indicator,
        config,
        battery_monitor,
    };
    Ok(Box::new(v))
}

impl BoardInteraction<'_> for V4<'_> {
    fn get_esp(&mut self) -> &mut ESP<'_> {
        &mut self.esp
    }

    fn get_config(&mut self) -> &PlantControllerConfig {
        &self.config
    }

    fn get_battery_monitor(&mut self) -> &mut Box<dyn BatteryInteraction> {
        &mut self.battery_monitor
    }

    fn set_charge_indicator(&mut self, charging: bool) -> anyhow::Result<()> {
        self.charge_indicator.set_state(charging.into()).expect("cannot fail");
        Ok(())
    }

    fn deep_sleep(&mut self, duration_in_ms: u64) -> ! {
        self.awake.set_low().unwrap();
        deep_sleep(duration_in_ms);
    }

    fn get_backup_info(&mut self) -> anyhow::Result<BackupHeader> {
        let store = bincode::serialize(&BackupHeader::default())?.len();
        let mut header_page_buffer = vec![0_u8; store];

        self.eeprom
            .read_data(0, &mut header_page_buffer)
            .map_err(|err| anyhow!("Error reading eeprom header {:?}", err))?;

        println!("Raw header is {:?} with size {}", header_page_buffer, store);
        let header: BackupHeader = bincode::deserialize(&header_page_buffer)?;
        anyhow::Ok(header)
    }

    fn get_backup_config(&mut self) -> anyhow::Result<Vec<u8>> {
        let store = bincode::serialize(&BackupHeader::default())?.len();
        let mut header_page_buffer = vec![0_u8; store];

        self.eeprom
            .read_data(0, &mut header_page_buffer)
            .map_err(|err| anyhow!("Error reading eeprom header {:?}", err))?;

        let header: BackupHeader = bincode::deserialize(&header_page_buffer)?;

        //skip page 0, used by the header
        let data_start_address = self.eeprom.page_size() as u32;
        let mut data_buffer = vec![0_u8; header.size];
        self.eeprom
            .read_data(data_start_address, &mut data_buffer)
            .map_err(|err| anyhow!("Error reading eeprom data {:?}", err))?;

        let checksum = X25.checksum(&data_buffer);
        if checksum != header.crc16 {
            bail!(
                "Invalid checksum, got {} but expected {}",
                checksum,
                header.crc16
            );
        }

        anyhow::Ok(data_buffer)
    }

    fn backup_config(&mut self, bytes: &[u8]) -> anyhow::Result<()> {
        let time = self.get_rtc_time()?.timestamp_millis();

        let delay = Delay::new_default();

        let checksum = X25.checksum(bytes);
        let page_size = self.eeprom.page_size();

        let header = BackupHeader {
            crc16: checksum,
            timestamp: time,
            size: bytes.len(),
        };

        let encoded = bincode::serialize(&header)?;
        if encoded.len() > page_size {
            bail!(
                "Size limit reached header is {}, but firest page is only {}",
                encoded.len(),
                page_size
            )
        }
        let as_u8: &[u8] = &encoded;

        match self.eeprom.write_page(0, as_u8) {
            OkStd(_) => {}
            Err(err) => bail!("Error writing eeprom {:?}", err),
        };
        delay.delay_ms(5);

        let to_write = bytes.chunks(page_size);

        let mut lastiter = 0;
        let mut current_page = 1;
        for chunk in to_write {
            let address = current_page * page_size as u32;
            self.eeprom
                .write_page(address, chunk)
                .map_err(|err| anyhow!("Error writing eeprom {:?}", err))?;
            current_page += 1;

            let iter = (current_page % 8) as usize;
            if iter != lastiter {
                for i in 0..PLANT_COUNT {
                    let _ = self.fault(i, iter == i);
                }
                lastiter = iter;
            }

            delay.delay_ms(5);
        }
        anyhow::Ok(())
    }

    fn is_day(&self) -> bool {
        self.solar_is_day.get_level().into()
    }

    fn water_temperature_c(&mut self) -> anyhow::Result<f32> {
        self.one_wire_bus
            .reset(&mut self.esp.delay)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;
        let first = self.one_wire_bus.devices(false, &mut self.esp.delay).next();
        if first.is_none() {
            bail!("Not found any one wire  Ds18b20");
        }
        let device_address = first
            .unwrap()
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;

        let water_temp_sensor = Ds18b20::new::<EspError>(device_address)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;

        water_temp_sensor
            .start_temp_measurement(&mut self.one_wire_bus, &mut self.esp.delay)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;
        ds18b20::Resolution::Bits12.delay_for_measurement_time(&mut self.esp.delay);
        let sensor_data = water_temp_sensor
            .read_data(&mut self.one_wire_bus, &mut self.esp.delay)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;
        if sensor_data.temperature == 85_f32 {
            bail!("Ds18b20 dummy temperature returned");
        }
        anyhow::Ok(sensor_data.temperature / 10_f32)
    }

    fn tank_sensor_voltage(&mut self) -> anyhow::Result<f32> {
        self.tank_power.set_high()?;
        //let stabilize
        self.esp.delay.delay_ms(100);

        let mut store = [0_u16; TANK_MULTI_SAMPLE];
        for multisample in 0..TANK_MULTI_SAMPLE {
            let value = self.tank_channel.read()?;
            store[multisample] = value;
        }
        self.tank_power.set_low()?;

        store.sort();
        let median_mv = store[6] as f32 / 1000_f32;
        anyhow::Ok(median_mv)
    }

    fn light(&mut self, enable: bool) -> anyhow::Result<()> {
        unsafe { gpio_hold_dis(self.light.pin()) };
        self.light.set_state(enable.into())?;
        unsafe { gpio_hold_en(self.light.pin()) };
        anyhow::Ok(())
    }

    fn pump(&mut self, plant: usize, enable: bool) -> anyhow::Result<()> {
        if enable {
            self.pump_expander.pin_set_high(GPIOBank::Bank0, plant.try_into()?)?;
        } else {
            self.pump_expander.pin_set_low(GPIOBank::Bank0, plant.try_into()?)?;
        }
        anyhow::Ok(())
    }

    fn fault(&mut self, plant: usize, enable: bool) -> anyhow::Result<()> {
        if enable {
            self.pump_expander.pin_set_high(GPIOBank::Bank1, plant.try_into()?)?
        } else {
            self.pump_expander.pin_set_low(GPIOBank::Bank1, plant.try_into()?)?
        }
        anyhow::Ok(())
    }

    fn measure_moisture_hz(&mut self, plant: usize, sensor: Sensor) -> anyhow::Result<f32> {
        let mut results = [0_f32; REPEAT_MOIST_MEASURE];
        for repeat in 0..REPEAT_MOIST_MEASURE {
            self.signal_counter.counter_pause()?;
            self.signal_counter.counter_clear()?;


            //Disable all
            self.sensor_expander.pin_set_high(GPIOBank::Bank0, MS4)?;

            let sensor_channel = match sensor {
                Sensor::A => plant as u32,
                Sensor::B => (15 - plant) as u32,
            };

            let is_bit_set = |b: u8| -> bool { sensor_channel & (1 << b) != 0 };
            if is_bit_set(0) {
                self.sensor_expander.pin_set_high(GPIOBank::Bank0, MS0)?;
            } else {
                self.sensor_expander.pin_set_low(GPIOBank::Bank0, MS0)?;
            }
            if is_bit_set(1) {
                self.sensor_expander.pin_set_high(GPIOBank::Bank0, MS1)?;
            } else {
                self.sensor_expander.pin_set_low(GPIOBank::Bank0, MS1)?;
            }
            if is_bit_set(2) {
                self.sensor_expander.pin_set_high(GPIOBank::Bank0, MS2)?;
            } else {
                self.sensor_expander.pin_set_low(GPIOBank::Bank0, MS2)?;
            }
            if is_bit_set(3) {
                self.sensor_expander.pin_set_high(GPIOBank::Bank0, MS3)?;
            } else {
                self.sensor_expander.pin_set_low(GPIOBank::Bank0, MS3)?;
            }

            self.sensor_expander.pin_set_low(GPIOBank::Bank0, MS4)?;
            self.sensor_expander.pin_set_high(GPIOBank::Bank0, SENSOR_ON)?;

            let delay = Delay::new_default();
            let measurement = 100; // TODO what is this scaling factor? what is its purpose?
            let factor = 1000f32 / measurement as f32;

            //give some time to stabilize
            delay.delay_ms(10);
            self.signal_counter.counter_resume()?;
            delay.delay_ms(measurement);
            self.signal_counter.counter_pause()?;
            self.sensor_expander.pin_set_high(GPIOBank::Bank0, MS4)?;
            self.sensor_expander.pin_set_low(GPIOBank::Bank0, SENSOR_ON)?;
            self.sensor_expander.pin_set_low(GPIOBank::Bank0, MS0)?;
            self.sensor_expander.pin_set_low(GPIOBank::Bank0, MS1)?;
            self.sensor_expander.pin_set_low(GPIOBank::Bank0, MS2)?;
            self.sensor_expander.pin_set_low(GPIOBank::Bank0, MS3)?;
            delay.delay_ms(10);
            let unscaled = self.signal_counter.get_counter_value()? as i32;
            let hz = unscaled as f32 * factor;
            log(
                LogMessage::RawMeasure,
                unscaled as u32,
                hz as u32,
                &plant.to_string(),
                &format!("{sensor:?}"),
            );
            results[repeat] = hz;
        }
        results.sort_by(|a, b| a.partial_cmp(b).unwrap()); // floats don't seem to implement total_ord

        let mid = results.len() / 2;
        let median = results[mid];
        anyhow::Ok(median)
    }

    fn general_fault(&mut self, enable: bool) {
        unsafe { gpio_hold_dis(self.general_fault.pin()) };
        self.general_fault.set_state(enable.into()).unwrap();
        unsafe { gpio_hold_en(self.general_fault.pin()) };
    }

    fn factory_reset(&mut self) -> anyhow::Result<()> {
        println!("factory resetting");
        self.esp.delete_config()?;
        //destroy backup header
        let dummy: [u8; 0] = [];
        self.backup_config(&dummy)?;

        anyhow::Ok(())
    }

    fn get_rtc_time(&mut self) -> anyhow::Result<DateTime<Utc>> {
        match self.rtc.datetime() {
            OkStd(rtc_time) => anyhow::Ok(rtc_time.and_utc()),
            Err(err) => {
                bail!("Error getting rtc time {:?}", err)
            }
        }
    }

    fn set_rtc_time(&mut self, time: &DateTime<Utc>) -> anyhow::Result<()> {
        let naive_time = time.naive_utc();
        match self.rtc.set_datetime(&naive_time) {
            OkStd(_) => anyhow::Ok(()),
            Err(err) => {
                bail!("Error getting rtc time {:?}", err)
            }
        }
    }

    fn test_pump(&mut self, plant: usize) -> anyhow::Result<()> {
        self.pump(plant, true)?;
        self.esp.delay.delay_ms(30000);
        self.pump(plant, false)?;
        anyhow::Ok(())
    }

    fn test(&mut self) -> anyhow::Result<()> {
        self.general_fault(true);
        self.esp.delay.delay_ms(100);
        self.general_fault(false);
        self.esp.delay.delay_ms(500);
        self.light(true)?;
        self.esp.delay.delay_ms(500);
        self.light(false)?;
        self.esp.delay.delay_ms(500);
        for i in 0..PLANT_COUNT {
            self.fault(i, true)?;
            self.esp.delay.delay_ms(500);
            self.fault(i, false)?;
            self.esp.delay.delay_ms(500);
        }
        for i in 0..PLANT_COUNT {
            self.pump(i, true)?;
            self.esp.delay.delay_ms(100);
            self.pump(i, false)?;
            self.esp.delay.delay_ms(100);
        }
        for plant in 0..PLANT_COUNT {
            let a = self.measure_moisture_hz(plant, Sensor::A);
            let b = self.measure_moisture_hz(plant, Sensor::B);
            let aa = match a {
                OkStd(a) => a as u32,
                Err(_) => u32::MAX,
            };
            let bb = match b {
                OkStd(b) => b as u32,
                Err(_) => u32::MAX,
            };
            log(LogMessage::TestSensor, aa, bb, &plant.to_string(), "");
        }
        self.esp.delay.delay_ms(10);
        anyhow::Ok(())
    }
}
